from typing import Any

from aiohttp import ClientSession

from garlandtools.models.lang import Lang
from garlandtools.models.records.base_record import BaseRecord
from garlandtools.models.records.factory import partial_factory, record_factory
from garlandtools.models.type import Type


class Client:
    base_url = "https://garlandtools.org/api/"

    def __init__(
        self,
        session: ClientSession | None = None,
        lang: Lang = Lang.EN,
    ):
        self._session = session
        self.lang = lang

    @property
    def session(self) -> ClientSession:
        if self._session is None or self._session.closed:
            self._session = ClientSession()
        return self._session

    async def search(
        self, query: str, type: Type | None = None, exact: bool = False
    ) -> list[BaseRecord]:
        """Search Garland Tools for the given query.

        Filters are not yet implemented.

        :param query: The query to search for.
        :param type: The record type you want to search for.
        :param exact: If the query should be an exact match.
        """
        params = {"text": query}
        if type is not None:
            params["type"] = type.value
        if exact:
            params["exact"] = "true"
        result = await self._get("search.php", params=params)

        return [partial_factory(r, client=self) for r in result]

    async def _get_by_id(self, id: int, type: Type):
        params = {
            "id": id,
            "type": type.value,
            "lang": self.lang.value,
            "version": 3 if type == type.LEVE or type == type.ITEM else 2,
        }
        return await self._get("get.php", params=params)

    async def get_by_id(self, id: int, type: Type):
        """Get a record by its ID.

        :param id: The ID of the record.
        :param type: The type of the record.
        :param lang: The language of the record.
        """
        result = await self._get_by_id(id, type)
        return record_factory(type, result, self)

    async def _get(self, path: str, params: dict | None = None) -> Any:
        url = self.base_url + path
        async with self.session.get(url, params=params) as response:
            if response.status == 200:
                return await response.json()
            else:
                raise Exception(f"Received status code {response.status} from {url}")



from typing import TYPE_CHECKING

from garlandtools.models.partials.item_partial import ItemPartial
from garlandtools.models.type import Type

if TYPE_CHECKING:
    from garlandtools.client import Client
    from garlandtools.models.records.base_record import BaseRecord


def partial_factory(partial_data: dict, client: "Client") -> "BaseRecord":  # type: ignore # noqa: E501
    partial_type: Type = Type(partial_data["type"])
    match partial_type:
        case Type.ITEM:
            from garlandtools.models.records.item import Item

            return Item(client=client, partial=ItemPartial(partial_data["obj"]))
        case _:
            # raise ValueError(f"Unknown type {partial_type}")
            print(f"Unknown type {partial_type}")


def record_factory(record_type: Type, record_data: dict, client: "Client"):
    record = record_data[record_type.value]
    related_records = [
        partial_factory(partial, client=client) for partial in record_data["partials"]
    ]
    match record_type:
        case Type.ITEM:
            from garlandtools.models.records.item import Item

            return Item(client, data=record, related_records=related_records)
        case _:
            raise ValueError(f"Unknown type {record_type}")



from garlandtools.models.records.base_record import BaseRecord
from garlandtools.models.type import Type


class Item(BaseRecord):
    @property
    def TYPE(self) -> Type:
        """The type of the record."""
        return Type.ITEM

    @property
    def ilvl(self) -> int:
        """The item level of the item."""
        if self.data is not None:
            return self.data["ilvl"]
        if self.partial is not None:
            return self.partial.ilvl
        raise ValueError("Neither data nor partial is set.")

    @property
    def icon(self) -> str:
        """The icon of the item."""
        if self.data is not None:
            return self.data["icon"]
        if self.partial is not None:
            return self.partial.icon
        raise ValueError("Neither data nor partial is set.")

    @property
    def category(self) -> int:
        """The category of the item."""
        if self.data is not None:
            return self.data["category"]
        if self.partial is not None:
            return self.partial.category
        raise ValueError("Neither data nor partial is set.")

    @property
    def price(self) -> int:
        """The price of the item."""
        if self.data is not None:
            if "price" not in self.data:
                return 0
            return self.data["price"]
        if self.partial is not None:
            return self.partial.price
        raise ValueError("Neither data nor partial is set.")

    @property
    async def description(self) -> str | None:
        """The description of the item."""
        return await self._get("description")

    # @property
    # async def patch(self) -> int:
    #     return self.data["patch"]  # TODO: return a Patch object

    # @property
    # async def patchCategory(self) -> int:
    #     return self.data["patchCategory"]  # TODO: return a PatchCategory object

    # @property
    # async def tradeable(self) -> bool:
    #     return bool(self.data["tradeable"])

    # @property
    # async def sell_price(self) -> int:
    #     return self.data["sell_price"]

    # @property
    # async def rarity(self) -> int:
    #     return self.data["rarity"]  # TODO: return a Rarity Enum

    @property
    async def stack_size(self) -> int | None:
        """The maximum number of items that can be stacked together."""
        return await self._get("stackSize")

    # @property
    # def nodes(self):
    #     return self.data['nodes']
    # @property
    # def vendors(self):
    #     return self.data['vendors']
    # @property
    # def ingredient_of(self):
    #     return self.data['ingredient_of']
    # @property
    # def leves(self):
    #     return self.data['leves']
    # @property
    # def ventures(self):
    #     return self.data['ventures']



class Partial:
    def __init__(self, data: dict):
        self.data = data

    @property
    def id(self) -> int:
        """The ID of the partial record."""
        return self.data["i"]

    @property
    def name(self) -> str:
        """The name of the partial record."""
        return self.data["n"]



from garlandtools.models.partials.partial import Partial


class ItemPartial(Partial):
    @property
    def ilvl(self) -> int:
        """The item level of the item."""
        return self.data["l"]

    @property
    def icon(self) -> str:
        """The icon of the item."""
        return self.data["c"]

    @property
    def category(self) -> int:
        """The category of the item."""
        return self.data["t"]

    @property
    def price(self) -> int:
        """The price of the item."""
        if "p" not in self.data:
            return 0
        return self.data["p"]



from enum import Enum


class Type(Enum):
    ACHIEVEMENT = "achievement"
    ACTION = "action"
    BROWSE = "browse"
    FATE = "fate"
    FISHING = "fishing"
    ITEM = "item"
    LEVE = "leve"
    MAP = "map"
    MOB = "mob"
    NODE = "node"
    NPC = "npc"
    QUEST = "quest"
    STATUS = "status"
    VENTURE = "venture"



import asyncio
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from garlandtools.models.records.factory import partial_factory
from garlandtools.models.type import Type

if TYPE_CHECKING:
    from garlandtools.client import Client

from garlandtools.models.partials.item_partial import ItemPartial


class BaseRecord(ABC):
    @property
    @abstractmethod
    def TYPE(self) -> Type:
        """The type of the record."""

    @property
    def id(self) -> int:
        """The ID of the item record."""
        if self.data is not None:
            return self.data["id"]
        if self.partial is not None:
            return self.partial.id
        raise ValueError("Neither data nor partial is set.")

    @property
    def name(self) -> str:
        """The name of the item record."""
        if self.data is not None:
            return self.data["name"]
        if self.partial is not None:
            return self.partial.name
        raise ValueError("Neither data nor partial is set.")

    def __init__(
        self,
        client: "Client",
        data: dict | None = None,
        partial: ItemPartial | None = None,
        related_records: list["BaseRecord"] = [],
    ):
        if data is None and partial is None:
            raise ValueError("Either data or partial must be provided.")
        self.client = client
        self.data = data
        self.partial = partial
        self.related_records = related_records
        self.fetch_lock = asyncio.Lock()

    async def _get(self, key: str) -> Any | None:
        if self.data is not None:
            return self.data[key]

        async with self.fetch_lock:
            # Check again in case another coroutine
            # fetched the data while we were waiting
            if self.data is not None:
                return self.data[key]
            record_data = await self.client._get_by_id(self.id, self.TYPE)
            self.data = record_data[self.TYPE.value]
            self.related_records = [
                partial_factory(partial, client=self.client)
                for partial in record_data["partials"]
            ]
            return self.data[key]
